<?php
/**
 * @package  SeoLinks
 * @license  http://www.gnu.org/licenseses/gpl-3.0.html GNU/GPL
 *
 * Created by Oleg Micriucov for Joomla! 3.x, 4.x, 5.x
 * https://allforjoomla.com
 *
 */
defined('_JEXEC') or die(':)');

use Joomla\Registry\Registry;

class plgSystemSeolinks extends JPlugin {
    
    const CACHE_KEY = 'plg_system_seolinks';
    
    private $_v = null;

    public function onAfterDispatch() {
        $app = JFactory::getApplication();
        if ($app->getName() != 'site') {
            return;
        }
        $document = JFactory::getDocument();
        if ($document->getType() != 'html') {
            return;
        }
        $input = $app->input;
        $task = $input->get('task', '');
        $format = $input->get('format', '');
        $tmpl = $input->get('tmpl', '');
        if ($task == 'edit' || ($format != '' && $format != 'html') || $tmpl != '') {
            return;
        }
        $this->_loadLang();
        $debugMode = (int)$this->params->get('debug_mode', 0);
        
        if($debugMode){
            $app->enqueueMessage('<b>SeoLinks</b>: '.JText::sprintf('PLG_SYSTEM_SEOLINKS_CURRENT_PAGE_URIS', implode(', ',$this->getCurrentPageURLs())));
        }
        if ($this->isThisSkipPage()) {
            if($debugMode){
                $app->enqueueMessage('<b>SeoLinks</b>: '.JText::_('PLG_SYSTEM_SEOLINKS_SKIPPING_PAGE'));
            }
            return true;
        }

        $body = trim($document->getBuffer('component'));
        if ($body == '') {
            return;
        } else {
            try {
                $newBody = $this->processLinks($body);
            } catch (\Exception $e) {
                if($debugMode){
                    $app->enqueueMessage('<b>SeoLinks</b>: '.$e->getMessage(), 'error');
                }
            }
        }
        if (!is_string($newBody) || $newBody == '') {
            return false;
        }
        $document->setBuffer($newBody, 'component');
    }
    
    public function onExtensionAfterSave($context, $table, $isNew = true) {
        if ($isNew || $context != 'com_plugins.plugin' || $table->element != 'seolinks') {
            return;
        }
        $cache = $this->_getCache();
        $cache->clean();
    }

    private function processLinks($body) {
        static $linksOnPage;
        
        $app = JFactory::getApplication();
        $maxOnPage = (int) $this->params->get('max_on_page', 100);
        $debugMode = (int)$this->params->get('debug_mode', 0);

        if (is_null($linksOnPage)) {
            $linksOnPage = 0;
        }

        $links = $this->getLinks();

        if (count($links) == 0) {
            if($debugMode){
                throw new \Exception(JText::_('PLG_SYSTEM_SEOLINKS_NO_LINKS'));
            }
            return false;
        }
        if ($maxOnPage > 0) {
            $linksOnPage += substr_count($body, '<a ');
            if ($linksOnPage >= $maxOnPage) {
                if($debugMode){
                    throw new \Exception(JText::sprintf('PLG_SYSTEM_SEOLINKS_X_TOO_MUCH_LINKS_ON_PAGE', $linksOnPage, $maxOnPage));
                }
                return false;
            }
            else if($debugMode){
                $app->enqueueMessage('<b>SeoLinks</b>: '.JText::sprintf('PLG_SYSTEM_SEOLINKS_X_LINKS_ON_PAGE', $linksOnPage, $maxOnPage));
            }
        }
        $body = preg_replace_callback('~(<\!\-\-seoLinks skip\-\->)(.*?)(<\!\-\-\/seoLinks skip\-\->)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[2]) . '<:ZyX/>';
        }, $body);
        $body = preg_replace_callback('~(<script)(.*?)(<\/script>)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
        }, $body);
        $body = preg_replace_callback('~(<\!\-\-)(.*?)(\-\-\>)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
        }, $body);
        $body = preg_replace_callback('~(<style)(.*?)(<\/style>)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
        }, $body);
        $body = preg_replace_callback('~(<h[1-6])(.*?)(<\/h[1-6]>)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
        }, $body);
        $body = preg_replace_callback('~(<a)(.*?)(<\/a>)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
        }, $body);
        $body = preg_replace_callback('~(<[a-z])(.*?)(>)~si', function ($matches) {
            return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
        }, $body);
        
        for ($i = 0; $i < count($links); $i++) {
            if ($maxOnPage > 0 && $linksOnPage >= $maxOnPage) {
                break;
            }
            $link = $links[$i];
            if ($link['hasNum'] >= $link['maxNum'] || in_array($link['attribs']['href'], $this->getCurrentPageURLs())) {
                continue;
            }
            $replace = '<a';
            foreach ($link['attribs'] as $attrKey => $attrVal) {
                if (!empty($attrVal)) {
                    $replace .= ' ' . $attrKey . '="' . $attrVal . '"';
                }
            }
            $replace .= '>$2</a>';
            $replace = '$1' . $replace . '$3';
            $search = '~([\s\.\,\;\!\?\:\>\(\)\'\"\*\/«])(' . $link['token'] . ')([\*\/\'\"\(\)\<\s\.\,\;\!\?\:»])~siu';
            $body = preg_replace_callback('~(<a)(.*?)(?=<\/a>)(<\/a>)~si', function ($matches) {
                return '<:ZyX>' . plgSystemSeolinks::maskContent($matches[1] . $matches[2] . $matches[3]) . '<:ZyX/>';
            }, $body);
            $body = preg_replace($search, $replace, $body, $link['maxNum']);
            if ($body == '') {
                return false;
            }
            $placedLinks = substr_count($body, '<a ');
            $link['hasNum'] += $placedLinks;
            $linksOnPage += $placedLinks;
            $links[$i] = $link;
        }
        $body = preg_replace_callback('~<\:ZyX>(.*?)(?=<\:ZyX\/>)<\:ZyX\/>~si', function ($matches) {
            return plgSystemSeolinks::unmaskContent($matches[1]);
        }, $body);
        $body = preg_replace_callback('~<\:ZyX>(.*?)(?=<\:ZyX\/>)<\:ZyX\/>~si', function ($matches) {
            return plgSystemSeolinks::unmaskContent($matches[1]);
        }, $body);
        return $body;
    }

    private function getLinks() {
        static $links;
        if (is_array($links)) {
            return $links;
        }
        $links = array();
        $cacheID = 'links';
        $caching = (int)$this->params->get('caching', 0);
        if($caching==1){
            $cache = $this->_getCache();
            $cached = $cache->get($cacheID);
            if ($cached !== false && !empty($cached)) {
                return json_decode($cached, true);
            }
        }
        $linx = $this->params->get('linx', '');
        $tmp = explode("\n", $linx);
        if (!is_array($tmp)) {
            return false;
        }
        foreach ($tmp as $linkExpr) {
            $linkExpr = trim($linkExpr);
            if(empty($linkExpr)){
                continue;
            }
            $links[] = $this->parseLinkExpr($linkExpr);
        }
        if($caching==1){
            $cache->store(json_encode($links, JSON_HEX_APOS), $cacheID);
        }
        return $links;
    }

    private function parseLinkExpr($linkExpr) {
        $linxCSS = $this->params->get('linx_class', '');
        $numRepl = max((int) $this->params->get('links_amount', 1), 1);
        $linkNum = max(1, $numRepl);
        
        if (strpos($linkExpr, '=') === false) {
            throw new \Exception(JText::sprintf('PLG_SYSTEM_SEOLINKS_INVALID_LINK_EXPR', $linkExpr));
        }
        $linkParts = explode('=', $linkExpr, 2);
        $words = trim($linkParts[0]);
        $linkData = trim($linkParts[1]);
        if (empty($words) || empty($linkData)) {
            throw new \Exception(JText::sprintf('PLG_SYSTEM_SEOLINKS_INVALID_LINK_EXPR', $linkExpr));
        }
        $wordFormWildCards = array(
            '\\\.' => '#-#',
            '\\\*' => '#--#',
            '.' => '[a-zа-яіїєґ]?',
            '*' => '[a-zа-яіїєґ]*'
        );
        $clearRegs = array(
            '#-#' => '\\.',
            '#--#' => '\\*'
        );
        $words = str_replace(', ', ',', $words);
        $words = str_replace(' ,', ',', $words);
        $words = str_replace(',', '|', $words);
        $words = str_replace(array_keys($wordFormWildCards), array_values($wordFormWildCards), addslashes($words));
        $words = str_replace(array_keys($clearRegs), array_values($clearRegs), $words);

        if (preg_match('~\{[0-9]+\}$~i', $linkData, $mchs)) {
            $linkNum = max((int) @$mchs[1], $numRepl);
            $linkData = trim(preg_replace('~\{[0-9]+\}$~i', '', $linkData));
        }

        $link = [
            'token' => $words,
            'attribs' => [
                'href' => '',
                'class' => $linxCSS,
            ],
            'maxNum' => $linkNum,
            'hasNum' => 0
        ];
        
        $json = json_decode($linkData, true);
        if (is_array($json) && !empty($json['href'])) {
            $link['attribs'] = array_merge($link['attribs'], $json);
        } else {
            $link['attribs']['href'] = $linkData;
        }
        
        return $link;
    }

    private function isThisSkipPage() {
        static $result;
        if (is_bool($result)) {
            return $result;
        }
        $skipPages = trim($this->params->get('skip_pages', ''));
        $uri = JURI::getInstance();
        if ($skipPages != '') {
            $skipPagesArray = explode("\n", $skipPages);
            foreach ($skipPagesArray as $skipPage) {
                $skipPage = trim($skipPage);
                if ($skipPage == '') {
                    continue;
                }
                if (substr($skipPage, 0, 1) == '~') {
                    if (preg_match($skipPage, $uri->toString(['host', 'port', 'path', 'query']))) {
                        $result = true;
                        return true;
                    }
                } else if (in_array($skipPage, $this->getCurrentPageURLs())) {
                    $result = true;
                    return true;
                }
            }
        }
        $result = false;
        return $result;
    }
    
    private function getCurrentPageURLs(){
        static $urls;
        if (is_array($urls)) {
            return $urls;
        }
        $uri = JURI::getInstance();
        $URL_O = $uri->toString(['host', 'port', 'path', 'query']);
        $URL_H = preg_replace('~^www\.~','',$URL_O);
        $URL_P = $uri->toString(['path', 'query']);
        $URL_O2 = $_SERVER['HTTP_HOST'].$_SERVER['REQUEST_URI'];
        $URL_H2 = preg_replace('~^www\.~','',$URL_O2);
        $URL_P2 = $_SERVER['REQUEST_URI'];
        $urls = [
            'http://'.$URL_H,
            'https://'.$URL_H,
            $URL_H,
            'http://www.'.$URL_H,
            'https://www.'.$URL_H,
            'www.'.$URL_H,
            $URL_P,
            'http://'.$URL_H2,
            'https://'.$URL_H2,
            $URL_H2,
            'http://www.'.$URL_H2,
            'https://www.'.$URL_H2,
            'www.'.$URL_H2,
            $URL_P2,
        ];
        return $urls;
    }

    static function maskContent($txt) {
        $txt = str_replace("\'", "'", $txt);
        //base64_encode is used only to ancode the HTML so it won't be processed
        $result = base64_encode($txt);
        return $result;
    }

    static function unmaskContent($txt) {
        $result = base64_decode($txt);
        return $result;
    }

    private function _loadLang() {
        $lang = JFactory::getLanguage();
        $lang->load('plg_system_seolinks', JPATH_ADMINISTRATOR);
    }

    private function _isAdmin() {
        $app = JFactory::getApplication();
        return ((method_exists($app, 'isAdmin') && $app->isAdmin()) || (method_exists($app, 'isClient') && $app->isClient('administrator')));
    }

    private function _setResponseBody($html) {
        if (version_compare($this->_getCoreVersion(), '4.0.0', 'ge')) {
            $app = JFactory::getApplication();
            return $app->setBody($html);
        }
        JResponse::setBody($html);
    }

    private function _getResponseBody() {
        if (version_compare($this->_getCoreVersion(), '4.0.0', 'ge')) {
            $app = JFactory::getApplication();
            return $app->getBody();
        }
        return JResponse::getBody();
    }

    private function _getCoreVersion() {
        if ($this->_v === null) {
            $this->_v = new JVersion;
        }
        $v = $this->_v->getShortVersion();
        if (strpos($v, '-') !== false) {
            $v = explode('-', $v);
            $v = $v[0];
        }
        return $v;
    }
    
    public function _getCache() {
        $conf = JFactory::getConfig();
        $lang = JFactory::getLanguage();
        $cachePath = $conf->get('cache_path', '');
        $cacheOptions = array(
            'defaultgroup' => self::CACHE_KEY,
            'language' => $lang->getTag(),
            'lifetime' => 99999,
            'caching' => true
        );
        if(empty($cachePath) && version_compare($this->_getCoreVersion(), '4.0.0', 'lt')){
            $cachePath = JPATH_ROOT . '/cache';
        }
        if(!empty($cachePath)){
            $cacheOptions['cachebase'] = $cachePath;
        }
        $cache = JCache::getInstance('', $cacheOptions);
        return $cache;
    }

}
